# -*- coding: utf-8 -*-
# © Copyright EnterpriseDB UK Limited 2011-2025
#
# This file is part of Barman.
#
# Barman is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Barman is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Barman.  If not, see <http://www.gnu.org/licenses/>.

import errno
import io
import os
from abc import ABCMeta, abstractmethod

from barman.exceptions import ArchivalBackupException
from barman.utils import with_metaclass


class AnnotationManager(with_metaclass(ABCMeta)):
    """
    This abstract base class defines the AnnotationManager interface which provides
    methods for read, write and delete of annotations for a given backup.
    """

    @abstractmethod
    def put_annotation(self, backup_id, key, value):
        """Add an annotation"""

    @abstractmethod
    def get_annotation(self, backup_id, key):
        """Get the value of an annotation"""

    @abstractmethod
    def delete_annotation(self, backup_id, key):
        """Delete an annotation"""


class AnnotationManagerFile(AnnotationManager):
    def __init__(self, path, old_path=None):
        """
        Constructor for the file-based annotation manager.
        Should be initialised with the path to the barman base backup directory.

        :param str path: The path where the annotation file should be placed.
        :param str|None old_path: Optional path used to read annotations written
            before Barman 3.13.3.

        .. note:
            Starting from Barman 3.13.3, annotation files were moved out of the base
            backup directory and into a dedicated metadata directory. While this class
            is agnostic about what kind of annotation it stores, backwards compatibility
            is needed for users upgrading from previous versions who may have annotations
            stored in the legacy location. For that reason, the *old_path* parameter
            allows this class to read and migrate existing annotations from the old
            directory as to maintain backwards compatibility.
        """
        self.path = path
        self.old_path = old_path

    def _get_old_annotation_path(self, backup_id, key):
        """
        Builds the annotation path for the specified *backup_id* and annotation *key* for
        annotations created before Barman 3.13.3. Check the note on this class'
        constructor for more context.

        :param str backup_id: The backup ID.
        :param str key: The annotation file name.
        :returns str: The path to the annotation.
        """
        return "%s/%s/annotations/%s" % (self.old_path, backup_id, key)

    def _get_annotation_path(self, backup_id, key):
        """
        Builds the annotation path for the specified backup_id and annotation key.

        :param str backup_id: The backup ID.
        :param str key: The annotation file name.
        :returns str: The path to the annotation.
        """
        return "%s/%s-%s" % (self.path, backup_id, key)

    def _check_and_relocate_old_annotation(self, backup_id, key):
        """
        Check if the annotation exists in the old path, used before Barman 3.13.3,
        and relocate it to the new path as to maintain backwards compatibility.

        :param str backup_id: The backup ID.
        :param str key: The annotation file name.
        """
        if not self.old_path:
            return
        old_path = self._get_old_annotation_path(backup_id, key)
        new_path = self._get_annotation_path(backup_id, key)
        if os.path.exists(old_path):
            os.rename(old_path, new_path)

    def delete_annotation(self, backup_id, key):
        """
        Deletes an annotation from the filesystem for the specified backup_id and
        annotation key.
        """
        self._check_and_relocate_old_annotation(backup_id, key)
        annotation_path = self._get_annotation_path(backup_id, key)
        try:
            os.remove(annotation_path)
        except EnvironmentError as e:
            # For Python 2 compatibility we must check the error code directly
            # If the annotation doesn't exist then the failure to delete it is not an
            # error condition and we should not proceed to remove the annotations
            # directory
            if e.errno == errno.ENOENT:
                return
            else:
                raise
        try:
            os.rmdir(os.path.dirname(annotation_path))
        except EnvironmentError as e:
            # For Python 2 compatibility we must check the error code directly
            # If we couldn't remove the directory because it wasn't empty then we
            # do not consider it an error condition
            if e.errno != errno.ENOTEMPTY:
                raise

    def get_annotation(self, backup_id, key):
        """
        Reads the annotation `key` for the specified backup_id from the filesystem
        and returns the value.
        """
        self._check_and_relocate_old_annotation(backup_id, key)
        annotation_path = self._get_annotation_path(backup_id, key)
        try:
            with open(annotation_path, "r") as annotation_file:
                return annotation_file.read()
        except EnvironmentError as e:
            # For Python 2 compatibility we must check the error code directly
            # If the annotation doesn't exist then return None
            if e.errno != errno.ENOENT:
                raise

    def put_annotation(self, backup_id, key, value):
        """
        Writes the specified value for annotation `key` for the specified backup_id
        to the filesystem.
        """
        annotation_path = self._get_annotation_path(backup_id, key)
        try:
            os.makedirs(os.path.dirname(annotation_path))
        except EnvironmentError as e:
            # For Python 2 compatibility we must check the error code directly
            # If the directory already exists then it is not an error condition
            if e.errno != errno.EEXIST:
                raise
        with open(annotation_path, "w") as annotation_file:
            if value:
                annotation_file.write(value)


class AnnotationManagerCloud(AnnotationManager):
    def __init__(self, cloud_interface, server_name):
        """
        Constructor for the cloud-based annotation manager.
        Should be initialised with the CloudInterface and name of the server which
        was used to create the backups.
        """
        self.cloud_interface = cloud_interface
        self.server_name = server_name
        self.annotation_cache = None

    def _get_base_path(self):
        """
        Returns the base path to the cloud storage, accounting for the fact that
        CloudInterface.path may be None.
        """
        return self.cloud_interface.path and "%s/" % self.cloud_interface.path or ""

    def _get_annotation_path(self, backup_id, key):
        """
        Builds the full key to the annotation in cloud storage for the specified
        backup_id and annotation key.
        """
        return "%s%s/base/%s/annotations/%s" % (
            self._get_base_path(),
            self.server_name,
            backup_id,
            key,
        )

    def _populate_annotation_cache(self):
        """
        Build a cache of which annotations actually exist by walking the bucket.
        This allows us to optimize get_annotation by just checking a (backup_id,key)
        tuple here which is cheaper (in time and money) than going to the cloud
        every time.
        """
        self.annotation_cache = {}
        for object_key in self.cloud_interface.list_bucket(
            os.path.join(self._get_base_path(), self.server_name, "base") + "/",
            delimiter="",
        ):
            key_parts = object_key.split("/")
            if len(key_parts) > 3:
                if key_parts[-2] == "annotations":
                    backup_id = key_parts[-3]
                    annotation_key = key_parts[-1]
                    self.annotation_cache[(backup_id, annotation_key)] = True

    def delete_annotation(self, backup_id, key):
        """
        Deletes an annotation from cloud storage for the specified backup_id and
        annotation key.
        """
        annotation_path = self._get_annotation_path(backup_id, key)
        self.cloud_interface.delete_objects([annotation_path])

    def get_annotation(self, backup_id, key, use_cache=True):
        """
        Reads the annotation `key` for the specified backup_id from cloud storage
        and returns the value.

        The default behaviour is that, when it is first run, it populates a
        cache of the annotations which exist for each backup by walking the
        bucket. Subsequent operations can check that cache and avoid having to
        call remote_open if an annotation is not found in the cache.

        This optimises for the case where annotations are sparse and assumes the
        cost of walking the bucket is less than the cost of the remote_open calls
        which would not return a value.

        In cases where we do not want to walk the bucket up front then the caching
        can be disabled.
        """
        # Optimize for the most common case where there is no annotation
        if use_cache:
            if self.annotation_cache is None:
                self._populate_annotation_cache()
            if (
                self.annotation_cache is not None
                and (backup_id, key) not in self.annotation_cache
            ):
                return None
        # We either know there's an annotation or we haven't used the cache so read
        # it from the cloud
        annotation_path = self._get_annotation_path(backup_id, key)
        annotation_fileobj = self.cloud_interface.remote_open(annotation_path)
        if annotation_fileobj:
            with annotation_fileobj:
                annotation_bytes = annotation_fileobj.readline()
                return annotation_bytes.decode("utf-8")
        else:
            # We intentionally return None if remote_open found nothing
            return None

    def put_annotation(self, backup_id, key, value):
        """
        Writes the specified value for annotation `key` for the specified backup_id
        to cloud storage.
        """
        annotation_path = self._get_annotation_path(backup_id, key)
        self.cloud_interface.upload_fileobj(
            io.BytesIO(value.encode("utf-8")), annotation_path
        )


class KeepManager(with_metaclass(ABCMeta, object)):
    """Abstract base class which defines the KeepManager interface"""

    ANNOTATION_KEY = "keep"

    TARGET_FULL = "full"
    TARGET_STANDALONE = "standalone"

    supported_targets = (TARGET_FULL, TARGET_STANDALONE)

    @abstractmethod
    def should_keep_backup(self, backup_id):
        pass

    @abstractmethod
    def keep_backup(self, backup_id, target):
        pass

    @abstractmethod
    def get_keep_target(self, backup_id):
        pass

    @abstractmethod
    def release_keep(self, backup_id):
        pass


class KeepManagerMixin(KeepManager):
    """
    A Mixin which adds KeepManager functionality to its subclasses.

    Keep management is built on top of annotations and consists of the
    following functionality:

      - Determine whether a given backup is intended to be kept beyond its retention
        period.
      - Determine the intended recovery target for the archival backup.
      - Add and remove the keep annotation.

    The functionality is implemented as a Mixin so that it can be used to add
    keep management to the backup management class in barman (BackupManager)
    as well as its closest analog in barman-cloud (CloudBackupCatalog).
    """

    def __init__(self, *args, **kwargs):
        """
        Base constructor (Mixin pattern).

        kwargs must contain *either*:
          - A barman.server.Server object with the key `server`, *or*:
          - A CloudInterface object and a server name, keys `cloud_interface` and
            `server_name` respectively.

        """
        if "server" in kwargs:
            server = kwargs.pop("server")
            self.annotation_manager = AnnotationManagerFile(
                server.meta_directory, server.config.basebackups_directory
            )
        elif "cloud_interface" in kwargs:
            self.annotation_manager = AnnotationManagerCloud(
                kwargs.pop("cloud_interface"), kwargs.pop("server_name")
            )
        super(KeepManagerMixin, self).__init__(*args, **kwargs)

    def should_keep_backup(self, backup_id):
        """
        Returns True if the specified backup_id for this server has a keep annotation.
        False otherwise.
        """
        return (
            self.annotation_manager.get_annotation(backup_id, type(self).ANNOTATION_KEY)
            is not None
        )

    def keep_backup(self, backup_id, target):
        """
        Add a keep annotation for backup with ID backup_id with the specified
        recovery target.
        """
        if target not in KeepManagerMixin.supported_targets:
            raise ArchivalBackupException("Unsupported recovery target: %s" % target)
        self.annotation_manager.put_annotation(
            backup_id, type(self).ANNOTATION_KEY, target
        )

    def get_keep_target(self, backup_id):
        """Retrieve the intended recovery target"""
        return self.annotation_manager.get_annotation(
            backup_id, type(self).ANNOTATION_KEY
        )

    def release_keep(self, backup_id):
        """Release the keep annotation"""
        self.annotation_manager.delete_annotation(backup_id, type(self).ANNOTATION_KEY)


class KeepManagerMixinCloud(KeepManagerMixin):
    """
    A specialised KeepManager which allows the annotation caching optimization in
    the AnnotationManagerCloud backend to be optionally disabled.
    """

    def should_keep_backup(self, backup_id, use_cache=True):
        """
        Like KeepManagerMixinCloud.should_keep_backup but with the use_cache option.
        """
        return (
            self.annotation_manager.get_annotation(
                backup_id, type(self).ANNOTATION_KEY, use_cache=use_cache
            )
            is not None
        )

    def get_keep_target(self, backup_id, use_cache=True):
        """
        Like KeepManagerMixinCloud.get_keep_target but with the use_cache option.
        """
        return self.annotation_manager.get_annotation(
            backup_id, type(self).ANNOTATION_KEY, use_cache=use_cache
        )



# -*- coding: utf-8 -*-
# © Copyright EnterpriseDB UK Limited 2011-2025
#
# This file is part of Barman.
#
# Barman is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Barman is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Barman.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains a wrapper for shell commands
"""

from __future__ import print_function

import errno
import inspect
import logging
import os
import re
import select
import signal
import subprocess
import sys
import time
from distutils.version import LooseVersion as Version

import barman.utils
from barman.exceptions import (
    CommandException,
    CommandFailedException,
    CommandMaxRetryExceeded,
)

_logger = logging.getLogger(__name__)


class Handler:
    def __init__(self, logger, level, prefix=None):
        self.class_logger = logger
        self.level = level
        self.prefix = prefix

    def run(self, line):
        if line:
            if self.prefix:
                self.class_logger.log(self.level, "%s%s", self.prefix, line)
            else:
                self.class_logger.log(self.level, "%s", line)

    __call__ = run


class StreamLineProcessor(object):
    """
    Class deputed to reading lines from a file object, using a buffered read.

    NOTE: This class never call os.read() twice in a row. And is designed to
    work with the select.select() method.
    """

    def __init__(self, fobject, handler):
        """
        :param file fobject: The file that is being read
        :param callable handler: The function (taking only one unicode string
         argument) which will be called for every line
        """
        self._file = fobject
        self._handler = handler
        self._buf = ""

    def fileno(self):
        """
        Method used by select.select() to get the underlying file descriptor.

        :rtype: the underlying file descriptor
        """
        return self._file.fileno()

    def process(self):
        """
        Read the ready data from the stream and for each line found invoke the
        handler.

        :return bool: True when End Of File has been reached
        """
        data = os.read(self._file.fileno(), 4096)
        # If nothing has been read, we reached the EOF
        if not data:
            self._file.close()
            # Handle the last line (always incomplete, maybe empty)
            self._handler(self._buf)
            return True
        self._buf += data.decode("utf-8", "replace")
        # If no '\n' is present, we just read a part of a very long line.
        # Nothing to do at the moment.
        if "\n" not in self._buf:
            return False
        tmp = self._buf.split("\n")
        # Leave the remainder in self._buf
        self._buf = tmp[-1]
        # Call the handler for each complete line.
        lines = tmp[:-1]
        for line in lines:
            self._handler(line)
        return False


class Command(object):
    """
    Wrapper for a system command
    """

    def __init__(
        self,
        cmd,
        args=None,
        env_append=None,
        path=None,
        shell=False,
        check=False,
        allowed_retval=(0,),
        close_fds=True,
        out_handler=None,
        err_handler=None,
        retry_times=0,
        retry_sleep=0,
        retry_handler=None,
    ):
        """
        If the `args` argument is specified the arguments will be always added
        to the ones eventually passed with the actual invocation.

        If the `env_append` argument is present its content will be appended to
        the environment of every invocation.

        The subprocess output and error stream will be processed through
        the output and error handler, respectively defined through the
        `out_handler` and `err_handler` arguments. If not provided every line
        will be sent to the log respectively at INFO and WARNING level.

        The `out_handler` and the `err_handler` functions will be invoked with
        one single argument, which is a string containing the line that is
        being processed.

        If the `close_fds` argument is True, all file descriptors
        except 0, 1 and 2 will be closed before the child process is executed.

        If the `check` argument is True, the exit code will be checked
        against the `allowed_retval` list, raising a CommandFailedException if
        not in the list.

        If `retry_times` is greater than 0, when the execution of a command
        terminates with an error, it will be retried for
        a maximum of `retry_times` times, waiting for `retry_sleep` seconds
        between every attempt.

        Every time a command is retried the `retry_handler` is executed
        before running the command again. The retry_handler must be a callable
        that accepts the following fields:

         * the Command object
         * the arguments list
         * the keyword arguments dictionary
         * the number of the failed attempt
         * the exception containing the error

        An example of such a function is:

            > def retry_handler(command, args, kwargs, attempt, exc):
            >     print("Failed command!")

        Some of the keyword arguments can be specified both in the class
        constructor and during the method call. If specified in both places,
        the method arguments will take the precedence over
        the constructor arguments.

        :param str cmd: The command to execute
        :param list[str]|None args: List of additional arguments to append
        :param dict[str.str]|None env_append: additional environment variables
        :param str path: PATH to be used while searching for `cmd`
        :param bool shell: If true, use the shell instead of an "execve" call
        :param bool check: Raise a CommandFailedException if the exit code
            is not present in `allowed_retval`
        :param list[int] allowed_retval: List of exit codes considered as a
            successful termination.
        :param bool close_fds: If set, close all the extra file descriptors
        :param callable out_handler: handler for lines sent on stdout
        :param callable err_handler: handler for lines sent on stderr
        :param int retry_times: number of allowed retry attempts
        :param int retry_sleep: wait seconds between every retry
        :param callable retry_handler: handler invoked during a command retry
        """
        self.pipe = None
        self.cmd = cmd
        self.args = args if args is not None else []
        self.shell = shell
        self.close_fds = close_fds
        self.check = check
        self.allowed_retval = allowed_retval
        self.retry_times = retry_times
        self.retry_sleep = retry_sleep
        self.retry_handler = retry_handler
        self.path = path
        self.ret = None
        self.out = None
        self.err = None
        # If env_append has been provided use it or replace with an empty dict
        env_append = env_append or {}
        # If path has been provided, replace it in the environment
        if path:
            env_append["PATH"] = path
        # Find the absolute path to the command to execute
        if not self.shell:
            full_path = barman.utils.which(self.cmd, self.path)
            if not full_path:
                raise CommandFailedException("%s not in PATH" % self.cmd)
            self.cmd = full_path
        # If env_append contains anything, build an env dict to be used during
        # subprocess call, otherwise set it to None and let the subprocesses
        # inherit the parent environment
        if env_append:
            self.env = os.environ.copy()
            self.env.update(env_append)
        else:
            self.env = None
        # If an output handler has been provided use it, otherwise log the
        # stdout as INFO
        if out_handler:
            self.out_handler = out_handler
        else:
            self.out_handler = self.make_logging_handler(logging.DEBUG)
        # If an error handler has been provided use it, otherwise log the
        # stderr as WARNING
        if err_handler:
            self.err_handler = err_handler
        else:
            self.err_handler = self.make_logging_handler(logging.WARNING)

    @staticmethod
    def _restore_sigpipe():
        """restore default signal handler (http://bugs.python.org/issue1652)"""
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)  # pragma: no cover

    def __call__(self, *args, **kwargs):
        """
        Run the command and return the exit code.

        The output and error strings are not returned, but they can be accessed
        as attributes of the Command object, as well as the exit code.

        If `stdin` argument is specified, its content will be passed to the
        executed command through the standard input descriptor.

        If the `close_fds` argument is True, all file descriptors
        except 0, 1 and 2 will be closed before the child process is executed.

        If the `check` argument is True, the exit code will be checked
        against the `allowed_retval` list, raising a CommandFailedException if
        not in the list.

        Every keyword argument can be specified both in the class constructor
        and during the method call. If specified in both places,
        the method arguments will take the precedence over
        the constructor arguments.

        :rtype: int
        :raise: CommandFailedException
        :raise: CommandMaxRetryExceeded
        """
        self.get_output(*args, **kwargs)
        return self.ret

    def get_output(self, *args, **kwargs):
        """
        Run the command and return the output and the error as a tuple.

        The return code is not returned, but it can be accessed as an attribute
        of the Command object, as well as the output and the error strings.

        If `stdin` argument is specified, its content will be passed to the
        executed command through the standard input descriptor.

        If the `close_fds` argument is True, all file descriptors
        except 0, 1 and 2 will be closed before the child process is executed.

        If the `check` argument is True, the exit code will be checked
        against the `allowed_retval` list, raising a CommandFailedException if
        not in the list.

        Every keyword argument can be specified both in the class constructor
        and during the method call. If specified in both places,
        the method arguments will take the precedence over
        the constructor arguments.

        :rtype: tuple[str, str]
        :raise: CommandFailedException
        :raise: CommandMaxRetryExceeded
        """
        attempt = 0
        while True:
            try:
                return self._get_output_once(*args, **kwargs)
            except CommandFailedException as exc:
                # Try again if retry number is lower than the retry limit
                if attempt < self.retry_times:
                    # If a retry_handler is defined, invoke it passing the
                    # Command instance and the exception
                    if self.retry_handler:
                        self.retry_handler(self, args, kwargs, attempt, exc)
                    # Sleep for configured time, then try again
                    time.sleep(self.retry_sleep)
                    attempt += 1
                else:
                    if attempt == 0:
                        # No retry requested by the user
                        # Raise the original exception
                        raise
                    else:
                        # If the max number of attempts is reached and
                        # there is still an error, exit raising
                        # a CommandMaxRetryExceeded exception and wrap the
                        # original one
                        raise CommandMaxRetryExceeded(*exc.args)

    def _get_output_once(self, *args, **kwargs):
        """
        Run the command and return the output and the error as a tuple.

        The return code is not returned, but it can be accessed as an attribute
        of the Command object, as well as the output and the error strings.

        If `stdin` argument is specified, its content will be passed to the
        executed command through the standard input descriptor.

        If the `close_fds` argument is True, all file descriptors
        except 0, 1 and 2 will be closed before the child process is executed.

        If the `check` argument is True, the exit code will be checked
        against the `allowed_retval` list, raising a CommandFailedException if
        not in the list.

        Every keyword argument can be specified both in the class constructor
        and during the method call. If specified in both places,
        the method arguments will take the precedence over
        the constructor arguments.

        :rtype: tuple[str, str]
        :raises: CommandFailedException
        """
        out = []
        err = []

        def out_handler(line):
            out.append(line)
            if self.out_handler is not None:
                self.out_handler(line)

        def err_handler(line):
            err.append(line)
            if self.err_handler is not None:
                self.err_handler(line)

        # If check is true, it must be handled here
        check = kwargs.pop("check", self.check)
        allowed_retval = kwargs.pop("allowed_retval", self.allowed_retval)
        self.execute(
            out_handler=out_handler,
            err_handler=err_handler,
            check=False,
            *args,
            **kwargs
        )
        self.out = "\n".join(out)
        self.err = "\n".join(err)

        _logger.debug("Command stdout: %s", self.out)
        _logger.debug("Command stderr: %s", self.err)

        # Raise if check and the return code is not in the allowed list
        if check:
            self.check_return_value(allowed_retval)
        return self.out, self.err

    def check_return_value(self, allowed_retval):
        """
        Check the current return code and raise CommandFailedException when
        it's not in the allowed_retval list

        :param list[int] allowed_retval: list of return values considered
            success
        :raises: CommandFailedException
        """
        if self.ret not in allowed_retval:
            raise CommandFailedException(dict(ret=self.ret, out=self.out, err=self.err))

    def execute(self, *args, **kwargs):
        """
        Execute the command and pass the output to the configured handlers

        If `stdin` argument is specified, its content will be passed to the
        executed command through the standard input descriptor.

        The subprocess output and error stream will be processed through
        the output and error handler, respectively defined through the
        `out_handler` and `err_handler` arguments. If not provided every line
        will be sent to the log respectively at INFO and WARNING level.

        If the `close_fds` argument is True, all file descriptors
        except 0, 1 and 2 will be closed before the child process is executed.

        If the `check` argument is True, the exit code will be checked
        against the `allowed_retval` list, raising a CommandFailedException if
        not in the list.

        Every keyword argument can be specified both in the class constructor
        and during the method call. If specified in both places,
        the method arguments will take the precedence over
        the constructor arguments.

        :rtype: int
        :raise: CommandFailedException
        """
        # Check keyword arguments
        stdin = kwargs.pop("stdin", None)
        check = kwargs.pop("check", self.check)
        allowed_retval = kwargs.pop("allowed_retval", self.allowed_retval)
        close_fds = kwargs.pop("close_fds", self.close_fds)
        out_handler = kwargs.pop("out_handler", self.out_handler)
        err_handler = kwargs.pop("err_handler", self.err_handler)
        if len(kwargs):
            raise TypeError(
                "%s() got an unexpected keyword argument %r"
                % (inspect.stack()[1][3], kwargs.popitem()[0])
            )

        # Reset status
        self.ret = None
        self.out = None
        self.err = None

        # Create the subprocess and save it in the current object to be usable
        # by signal handlers
        pipe = self._build_pipe(args, close_fds)
        self.pipe = pipe

        # Send the provided input and close the stdin descriptor
        if stdin:
            pipe.stdin.write(stdin)
        pipe.stdin.close()
        # Prepare the list of processors
        processors = [
            StreamLineProcessor(pipe.stdout, out_handler),
            StreamLineProcessor(pipe.stderr, err_handler),
        ]

        # Read the streams until the subprocess exits
        self.pipe_processor_loop(processors)

        # Reap the zombie and read the exit code
        pipe.wait()
        self.ret = pipe.returncode

        # Remove the closed pipe from the object
        self.pipe = None
        _logger.debug("Command return code: %s", self.ret)

        # Raise if check and the return code is not in the allowed list
        if check:
            self.check_return_value(allowed_retval)
        return self.ret

    def _build_pipe(self, args, close_fds):
        """
        Build the Pipe object used by the Command

        The resulting command will be composed by:
           self.cmd + self.args + args

        :param args: extra arguments for the subprocess
        :param close_fds: if True all file descriptors except 0, 1 and 2
            will be closed before the child process is executed.
        :rtype: subprocess.Popen
        """
        # Append the argument provided to this method of the base argument list
        args = self.args + list(args)
        # If shell is True, properly quote the command
        if self.shell:
            cmd = full_command_quote(self.cmd, args)
        else:
            cmd = [self.cmd] + args

        # Log the command we are about to execute
        _logger.debug("Command: %r", cmd)
        return subprocess.Popen(
            cmd,
            shell=self.shell,
            env=self.env,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            preexec_fn=self._restore_sigpipe,
            close_fds=close_fds,
        )

    @staticmethod
    def pipe_processor_loop(processors):
        """
        Process the output received through the pipe until all the provided
        StreamLineProcessor reach the EOF.

        :param list[StreamLineProcessor] processors: a list of
            StreamLineProcessor
        """
        # Loop until all the streams reaches the EOF
        while processors:
            try:
                ready = select.select(processors, [], [])[0]
            except select.error as e:
                # If the select call has been interrupted by a signal
                # just retry
                if e.args[0] == errno.EINTR:
                    continue
                raise

            # For each ready StreamLineProcessor invoke the process() method
            for stream in ready:
                eof = stream.process()
                # Got EOF on this stream
                if eof:
                    # Remove the stream from the list of valid processors
                    processors.remove(stream)

    @classmethod
    def make_logging_handler(cls, level, prefix=None):
        """
        Build a handler function that logs every line it receives.

        The resulting callable object logs its input at the specified level
        with an optional prefix.

        :param level: The log level to use
        :param prefix: An optional prefix to prepend to the line
        :return: handler function
        """
        class_logger = logging.getLogger(cls.__name__)

        return Handler(class_logger, level, prefix)

    @staticmethod
    def make_output_handler(prefix=None):
        """
        Build a handler function which prints every line it receives.

        The resulting function prints (and log it at INFO level) its input
        with an optional prefix.

        :param prefix: An optional prefix to prepend to the line
        :return: handler function
        """

        # Import the output module inside the function to avoid circular
        # dependency
        from barman import output

        def handler(line):
            if line:
                if prefix:
                    output.info("%s%s", prefix, line)
                else:
                    output.info("%s", line)

        return handler

    def enable_signal_forwarding(self, signal_id):
        """
        Enable signal forwarding to the subprocess for a specified signal_id

        :param signal_id: The signal id to be forwarded
        """
        # Get the current signal handler
        old_handler = signal.getsignal(signal_id)

        def _handler(sig, frame):
            """
            This signal handler forward the signal to the subprocess then
            execute the original handler.
            """
            # Forward the signal to the subprocess
            if self.pipe:
                self.pipe.send_signal(signal_id)
            # If the old handler is callable
            if callable(old_handler):
                old_handler(sig, frame)
            # If we have got a SIGTERM, we must exit
            elif old_handler == signal.SIG_DFL and signal_id == signal.SIGTERM:
                sys.exit(128 + signal_id)

        # Set the signal handler
        signal.signal(signal_id, _handler)


class Rsync(Command):
    """
    This class is a wrapper for the rsync system command,
    which is used vastly by barman
    """

    def __init__(
        self,
        rsync="rsync",
        args=None,
        ssh=None,
        ssh_options=None,
        bwlimit=None,
        exclude=None,
        exclude_and_protect=None,
        include=None,
        network_compression=None,
        path=None,
        **kwargs
    ):
        """
        :param str rsync: rsync executable name
        :param list[str]|None args: List of additional argument to always
            append
        :param str ssh: the ssh executable to be used when building
            the `-e` argument
        :param list[str] ssh_options: the ssh options to be used when building
            the `-e` argument
        :param str bwlimit: optional bandwidth limit
        :param list[str] exclude: list of file to be excluded from the copy
        :param list[str] exclude_and_protect: list of file to be excluded from
            the copy, preserving the destination if exists
        :param list[str] include: list of files to be included in the copy
            even if excluded.
        :param bool network_compression: enable the network compression
        :param str path: PATH to be used while searching for `cmd`
        :param bool check: Raise a CommandFailedException if the exit code
            is not present in `allowed_retval`
        :param list[int] allowed_retval: List of exit codes considered as a
            successful termination.
        """
        options = []
        if ssh:
            options += ["-e", full_command_quote(ssh, ssh_options)]
        if network_compression:
            options += ["-z"]
        # Include patterns must be before the exclude ones, because the exclude
        # patterns actually short-circuit the directory traversal stage
        # when rsync finds the files to send.
        if include:
            for pattern in include:
                options += ["--include=%s" % (pattern,)]
        if exclude:
            for pattern in exclude:
                options += ["--exclude=%s" % (pattern,)]
        if exclude_and_protect:
            for pattern in exclude_and_protect:
                options += ["--exclude=%s" % (pattern,), "--filter=P_%s" % (pattern,)]
        if args:
            options += self._args_for_suse(args)
        if bwlimit is not None and bwlimit > 0:
            options += ["--bwlimit=%s" % bwlimit]

        # By default check is on and the allowed exit code are 0 and 24
        if "check" not in kwargs:
            kwargs["check"] = True
        if "allowed_retval" not in kwargs:
            kwargs["allowed_retval"] = (0, 24)
        Command.__init__(self, rsync, args=options, path=path, **kwargs)

    def _args_for_suse(self, args):
        """
        Mangle args for SUSE compatibility

        See https://bugzilla.opensuse.org/show_bug.cgi?id=898513
        """
        # Prepend any argument starting with ':' with a space
        # Workaround for SUSE rsync issue
        return [" " + a if a.startswith(":") else a for a in args]

    def get_output(self, *args, **kwargs):
        """
        Run the command and return the output and the error (if present)
        """
        # Prepares args for SUSE
        args = self._args_for_suse(args)
        # Invoke the base class method
        return super(Rsync, self).get_output(*args, **kwargs)

    def from_file_list(self, filelist, src, dst, *args, **kwargs):
        """
        This method copies filelist from src to dst.

        Returns the return code of the rsync command
        """
        if "stdin" in kwargs:
            raise TypeError("from_file_list() doesn't support 'stdin' keyword argument")
        # The input string for the rsync --files-from argument must have a
        # trailing newline for compatibility with certain versions of rsync.
        input_string = ("\n".join(filelist) + "\n").encode("UTF-8")
        _logger.debug("from_file_list: %r", filelist)
        kwargs["stdin"] = input_string
        self.get_output("--files-from=-", src, dst, *args, **kwargs)
        return self.ret


class RsyncPgData(Rsync):
    """
    This class is a wrapper for rsync, specialised in sync-ing the
    Postgres data directory
    """

    def __init__(self, rsync="rsync", args=None, **kwargs):
        """
        Constructor

        :param str rsync: command to run
        """
        options = ["-rLKpts", "--delete-excluded", "--inplace"]
        if args:
            options += args
        Rsync.__init__(self, rsync, args=options, **kwargs)


class PostgreSQLClient(Command):
    """
    Superclass of all the PostgreSQL client commands.
    """

    COMMAND_ALTERNATIVES = None
    """
    Sometimes the name of a command has been changed during the PostgreSQL
    evolution. I.e. that happened with pg_receivexlog, that has been renamed
    to pg_receivewal. In that case, we should try using pg_receivewal (the
    newer alternative) and, if that command doesn't exist, we should try
    using `pg_receivexlog`.

    This is a list of command names to be used to find the installed command.
    """

    def __init__(
        self, connection, command, version=None, app_name=None, path=None, **kwargs
    ):
        """
        Constructor

        :param PostgreSQL connection: an object representing
          a database connection
        :param str command: the command to use
        :param Version version: the command version
        :param str app_name: the application name to use for the connection
        :param str path: additional path for executable retrieval
        """
        Command.__init__(self, command, path=path, **kwargs)
        if not connection:
            self.enable_signal_forwarding(signal.SIGINT)
            self.enable_signal_forwarding(signal.SIGTERM)
            return
        if version and version >= Version("9.3"):
            # If version of the client is >= 9.3 we use the connection
            # string because allows the user to use all the parameters
            # supported by the libpq library to create a connection
            conn_string = connection.get_connection_string(app_name)
            self.args.append("--dbname=%s" % conn_string)
        else:
            # 9.2 version doesn't support
            # connection strings so the 'split' version of the conninfo
            # option is used instead.
            conn_params = connection.conn_parameters
            self.args.append("--host=%s" % conn_params.get("host", None))
            self.args.append("--port=%s" % conn_params.get("port", None))
            self.args.append("--username=%s" % conn_params.get("user", None))

        self.enable_signal_forwarding(signal.SIGINT)
        self.enable_signal_forwarding(signal.SIGTERM)

    @classmethod
    def find_command(cls, path=None):
        """
        Find the active command, given all the alternatives as set in the
        property named `COMMAND_ALTERNATIVES` in this class.

        :param str path: The path to use while searching for the command
        :rtype: Command
        """

        # TODO: Unit tests of this one

        # To search for an available command, testing if the command
        # exists in PATH is not sufficient. Debian will install wrappers for
        # all commands, even if the real command doesn't work.
        #
        # I.e. we may have a wrapper for `pg_receivewal` even it PostgreSQL
        # 10 isn't installed.
        #
        # This is an example of what can happen in this case:
        #
        # ```
        # $ pg_receivewal --version; echo $?
        # Error: pg_wrapper: pg_receivewal was not found in
        #   /usr/lib/postgresql/9.6/bin
        # 1
        # $ pg_receivexlog --version; echo $?
        # pg_receivexlog (PostgreSQL) 9.6.3
        # 0
        # ```
        #
        # That means we should not only ensure the existence of the command,
        # but we also need to invoke the command to see if it is a shim
        # or not.

        # Get the system path if needed
        if path is None:
            path = os.getenv("PATH")
        # If the path is None at this point we have nothing to search
        if path is None:
            path = ""

        # Search the requested executable in every directory present
        # in path and return a Command object first occurrence that exists,
        # is executable and runs without errors.
        for path_entry in path.split(os.path.pathsep):
            for cmd in cls.COMMAND_ALTERNATIVES:
                full_path = barman.utils.which(cmd, path_entry)

                # It doesn't exist try another
                if not full_path:
                    continue

                # It exists, let's try invoking it with `--version` to check if
                # it's real or not.
                try:
                    command = Command(full_path, path=path, check=True)
                    command("--version")
                    return command
                except CommandFailedException:
                    # It's only a inactive shim
                    continue

        # We don't have such a command
        raise CommandFailedException(
            "command not in PATH, tried: %s" % " ".join(cls.COMMAND_ALTERNATIVES)
        )

    @classmethod
    def get_version_info(cls, path=None):
        """
        Return a dictionary containing all the info about
        the version of the PostgreSQL client

        :param str path: the PATH env
        """
        if cls.COMMAND_ALTERNATIVES is None:
            raise NotImplementedError(
                "get_version_info cannot be invoked on %s" % cls.__name__
            )

        version_info = dict.fromkeys(
            ("full_path", "full_version", "major_version"), None
        )

        # Get the version string
        try:
            command = cls.find_command(path)
        except CommandFailedException as e:
            _logger.debug("Error invoking %s: %s", cls.__name__, e)
            return version_info

        version_info["full_path"] = command.cmd
        # Parse the full text version
        try:
            full_version = command.out.strip()
            # Remove values inside parenthesis, they
            # carries additional information we do not need.
            full_version = re.sub(r"\s*\([^)]*\)", "", full_version)
            full_version = full_version.split()[1]
        except IndexError:
            _logger.debug("Error parsing %s version output", version_info["full_path"])
            return version_info

        if not re.match(r"(\d+)(\.(\d+)|devel|beta|alpha|rc).*", full_version):
            _logger.debug("Error parsing %s version output", version_info["full_path"])
            return version_info

        # Extract the major version
        version_info["full_version"] = Version(full_version)
        version_info["major_version"] = Version(
            barman.utils.simplify_version(full_version)
        )

        return version_info


class PgBaseBackup(PostgreSQLClient):
    """
    Wrapper class for the pg_basebackup system command
    """

    COMMAND_ALTERNATIVES = ["pg_basebackup"]

    def __init__(
        self,
        connection,
        destination,
        command,
        version=None,
        app_name=None,
        bwlimit=None,
        tbs_mapping=None,
        immediate=False,
        check=True,
        compression=None,
        parent_backup_manifest_path=None,
        args=None,
        **kwargs
    ):
        """
        Constructor

        :param PostgreSQL connection: an object representing
          a database connection
        :param str destination: destination directory path
        :param str command: the command to use
        :param Version version: the command version
        :param str app_name: the application name to use for the connection
        :param str bwlimit: bandwidth limit for pg_basebackup
        :param Dict[str, str] tbs_mapping: used for tablespace
        :param bool immediate: fast checkpoint identifier for pg_basebackup
        :param bool check: check if the return value is in the list of
          allowed values of the Command obj
        :param barman.compression.PgBaseBackupCompression compression:
          the pg_basebackup compression options used for this backup
        :param str parent_backup_manifest_path:
          the path to a backup_manifest file from a previous backup which can
          be used to perform an incremental backup
        :param List[str] args: additional arguments
        """
        PostgreSQLClient.__init__(
            self,
            connection=connection,
            command=command,
            version=version,
            app_name=app_name,
            check=check,
            **kwargs
        )

        # Set the backup destination
        self.args += ["-v", "--no-password", "--pgdata=%s" % destination]

        if version and version >= Version("10"):
            # If version of the client is >= 10 it would use
            # a temporary replication slot by default to keep WALs.
            # We don't need it because Barman already stores the full
            # WAL stream, so we disable this feature to avoid wasting one slot.
            self.args += ["--no-slot"]
            # We also need to specify that we do not want to fetch any WAL file
            self.args += ["--wal-method=none"]

        # The tablespace mapping option is repeated once for each tablespace
        if tbs_mapping:
            for tbs_source, tbs_destination in tbs_mapping.items():
                self.args.append(
                    "--tablespace-mapping=%s=%s" % (tbs_source, tbs_destination)
                )

        # Only global bandwidth limit is supported
        if bwlimit is not None and bwlimit > 0:
            self.args.append("--max-rate=%s" % bwlimit)

        # If it has a manifest file path it means it is an incremental backup
        if parent_backup_manifest_path:
            self.args.append("--incremental=%s" % parent_backup_manifest_path)

        # Immediate checkpoint
        if immediate:
            self.args.append("--checkpoint=fast")

        # Append compression arguments, the exact format of which are determined
        # in another function since they depend on the command version
        self.args.extend(self._get_compression_args(version, compression))

        # Manage additional args
        if args:
            self.args += args

    def _get_compression_args(self, version, compression):
        """
        Determine compression related arguments for pg_basebackup from the supplied
        compression options in the format required by the pg_basebackup version.

        :param Version version: The pg_basebackup version for which the arguments
          should be formatted.
        :param barman.compression.PgBaseBackupCompression compression:
          the pg_basebackup compression options used for this backup
        """
        compression_args = []

        if compression is not None:
            if compression.config.format is not None:
                compression_format = compression.config.format
            else:
                compression_format = "tar"
            compression_args.append("--format=%s" % compression_format)
            # For clients >= 15 we use the new --compress argument format
            if version and version >= Version("15"):
                compress_arg = "--compress="
                detail = []
                if compression.config.location is not None:
                    compress_arg += "%s-" % compression.config.location
                compress_arg += compression.config.type
                if compression.config.level is not None:
                    detail.append("level=%d" % compression.config.level)
                if compression.config.workers is not None:
                    detail.append("workers=%d" % compression.config.workers)
                if detail:
                    compress_arg += ":%s" % ",".join(detail)

                compression_args.append(compress_arg)
            # For clients < 15 we use the old style argument format
            else:
                if compression.config.type == "none":
                    compression_args.append("--compress=0")
                else:
                    if compression.config.level is not None:
                        compression_args.append(
                            "--compress=%d" % compression.config.level
                        )
                    # --gzip must be positioned after --compress when compression level=0
                    # so `base.tar.gz` can be created. Otherwise `.gz` won't be added.
                    compression_args.append("--%s" % compression.config.type)
        return compression_args


class PgReceiveXlog(PostgreSQLClient):
    """
    Wrapper class for pg_receivexlog
    """

    COMMAND_ALTERNATIVES = ["pg_receivewal", "pg_receivexlog"]

    def __init__(
        self,
        connection,
        destination,
        command,
        version=None,
        app_name=None,
        synchronous=False,
        check=True,
        slot_name=None,
        args=None,
        **kwargs
    ):
        """
        Constructor

        :param PostgreSQL connection: an object representing
          a database connection
        :param str destination: destination directory path
        :param str command: the command to use
        :param Version version: the command version
        :param str app_name: the application name to use for the connection
        :param bool synchronous: request synchronous WAL streaming
        :param bool check: check if the return value is in the list of
          allowed values of the Command obj
        :param str slot_name: the replication slot name to use for the
          connection
        :param List[str] args: additional arguments
        """
        PostgreSQLClient.__init__(
            self,
            connection=connection,
            command=command,
            version=version,
            app_name=app_name,
            check=check,
            **kwargs
        )

        self.args += [
            "--verbose",
            "--no-loop",
            "--no-password",
            "--directory=%s" % destination,
        ]

        # Add the replication slot name if set in the configuration.
        if slot_name is not None:
            self.args.append("--slot=%s" % slot_name)
        # Request synchronous mode
        if synchronous:
            self.args.append("--synchronous")

        # Manage additional args
        if args:
            self.args += args


class PgVerifyBackup(PostgreSQLClient):
    """
    Wrapper class for the pg_verify system command
    """

    COMMAND_ALTERNATIVES = ["pg_verifybackup"]

    def __init__(
        self,
        data_path,
        command,
        connection=None,
        version=None,
        app_name=None,
        check=True,
        args=None,
        **kwargs
    ):
        """
        Constructor

        :param str data_path: backup data directory
        :param str command: the command to use
        :param PostgreSQL connection: an object representing a database connection
        :param Version version: the command version
        :param str app_name: the application name to use for the connection
        :param bool check: check if the return value is in the list of allowed values of
            the Command obj
        :param List[str] args: additional arguments
        """
        PostgreSQLClient.__init__(
            self,
            connection=connection,
            command=command,
            version=version,
            app_name=app_name,
            check=check,
            **kwargs
        )

        self.args = ["-n", data_path]
        if args:
            self.args += args


class PgCombineBackup(PostgreSQLClient):
    """
    Wrapper class for the ``pg_combinebackup`` system command
    """

    COMMAND_ALTERNATIVES = ["pg_combinebackup"]

    def __init__(
        self,
        destination,
        command,
        tbs_mapping=None,
        connection=None,
        version=None,
        app_name=None,
        check=True,
        args=None,
        **kwargs
    ):
        """
        Constructor

        :param str destination: destination directory path
        :param str command: the command to use
        :param None|Dict[str, str] tbs_mapping: used for tablespace
        :param PostgreSQL connection: an object representing
          a database connection
        :param Version version: the command version
        :param str app_name: the application name to use for the connection
        :param bool check: check if the return value is in the list of
          allowed values of the :class:`Command` obj
        :param None|List[str] args: additional arguments
        """
        PostgreSQLClient.__init__(
            self,
            connection=connection,
            command=command,
            version=version,
            app_name=app_name,
            check=check,
            **kwargs
        )

        # Set the backup destination
        self.args = ["--output=%s" % destination]

        # The tablespace mapping option is repeated once for each tablespace
        if tbs_mapping:
            for tbs_source, tbs_destination in tbs_mapping.items():
                self.args.append(
                    "--tablespace-mapping=%s=%s" % (tbs_source, tbs_destination)
                )

        # Manage additional args
        if args:
            self.args += args


class BarmanSubProcess(object):
    """
    Wrapper class for barman sub instances
    """

    def __init__(
        self,
        command=sys.argv[0],
        subcommand=None,
        config=None,
        args=None,
        keep_descriptors=False,
    ):
        """
        Build a specific wrapper for all the barman sub-commands,
        providing a unified interface.

        :param str command: path to barman
        :param str subcommand: the barman sub-command
        :param str config: path to the barman configuration file.
        :param list[str] args: a list containing the sub-command args
            like the target server name
        :param bool keep_descriptors: whether to keep the subprocess stdin,
            stdout, stderr descriptors attached. Defaults to False
        """
        # The config argument is needed when the user explicitly
        # passes a configuration file, as the child process
        # must know the configuration file to use.
        #
        # The configuration file must always be propagated,
        # even in case of the default one.
        if not config:
            raise CommandFailedException(
                "No configuration file passed to barman subprocess"
            )
        # Build the sub-command:
        # * be sure to run it with the right python interpreter
        # * pass the current configuration file with -c
        # * set it quiet with -q
        self.command = [sys.executable, command, "-c", config, "-q", subcommand]
        self.keep_descriptors = keep_descriptors
        # Handle args for the sub-command (like the server name)
        if args:
            self.command += args

    def execute(self):
        """
        Execute the command and pass the output to the configured handlers
        """
        _logger.debug("BarmanSubProcess: %r", self.command)
        # Redirect all descriptors to /dev/null
        devnull = open(os.devnull, "a+")

        additional_arguments = {}
        if not self.keep_descriptors:
            additional_arguments = {"stdout": devnull, "stderr": devnull}

        proc = subprocess.Popen(
            self.command,
            preexec_fn=os.setsid,
            close_fds=True,
            stdin=devnull,
            **additional_arguments
        )
        _logger.debug("BarmanSubProcess: subprocess started. pid: %s", proc.pid)


class GPG(Command):
    """
    Wrapper class for the Gnu Privacy Guard (GPG) command-line tool.

    This class provides an interface to encrypt and decrypt data using GPG.

    .. note::
        GPG is a complete and free implementation of the OpenPGP standard as
        defined by RFC4880.

        It supports encryption, signing, and key management, and is designed for
        easy integration with other applications.

        In this class, we are only interested in encryption/decryption.

    .. important::
        `--pinentry-mode loopback` requires GPG version >= 2.1.

    .. important::
        This class expects sensitive data such as a GPG passphrase to be passed
        as a ``bytearray``, not as a ``str``. Python strings are immutable and may
        linger in memory, potentially exposing sensitive data. Using a ``bytearray``
        allows the caller to explicitly zero out the passphrase after use, similar
        to ``memset`` in C.

        Example:
            passphrase = bytearray(b"your-secret-passphrase")
            gpg(stdin=passphrase)
            passphrase[:] = b"\x00" * len(passphrase)

    Example for decryption:
        .. code-block:: python

                >>> gpg = GPG(
                ...     action="decrypt",
                ...     input_filepath="file.gpg",
                ...     output_filepath="file.decrypted"
                ... )
                >>> passphrase = bytearray(b"secret")
                >>> gpg(stdin=passphrase)
                >>> passphrase[:] = b"\x00" * len(passphrase)

    Example for decryption:
        .. code-block:: python

                >>> gpg = GPG(
                ...     action="encrypt",
                ...     input_filepath="file.txt",
                ...     output_filepath="file.txt.gpg",
                ...     recipient="user@example.com"
                ... )
                >>> gpg()
    """

    def __init__(
        self,
        gpg="gpg",
        action=None,
        recipient=None,
        input_filepath=None,
        output_filepath=None,
        **kwargs
    ):
        """
        Initialize the GPG command wrapper.

        :param str gpg: Path or name of the GPG executable. Defaults to ``gpg``.
        :param str action: Action to perform: ``encrypt`` or ``decrypt``.
        :param str recipient: Key identifier for encryption (required if *action* is
            ``encrypt``).
        :param str input_filepath: File to encrypt or decrypt.
        :param str output_filepath: Output file path for encrypted/decrypted data.
        :param kwargs: Additional keyword arguments passed to the base class:`Command`
            class.

        :raises CommandException: If ``encrypt`` is specified without a recipient.
        :raises ValueError: If *action* is invalid.
        """
        # Automatically answer "yes" to most prompts.
        # Use batch mode to suppress interactive prompts.
        # Set the pinentry mode to loopback for non-interactive passphrase entry.
        options = ["--yes", "--batch", "--pinentry-mode", "loopback"]

        if action == "decrypt":
            # Provide the passphrase via standard input (file descriptor 0) for
            # decryption.
            options += ["--passphrase-fd", "0", "--decrypt"]
        elif action == "encrypt":
            if not recipient:
                raise CommandException(
                    "A recipient must be specified to encrypt the backup. Please "
                    "provide a valid recipient."
                )
            # Set compression level to 0 (no compression) and specify recipient for
            # encryption.
            options += ["--compress-level", "0", "--recipient", recipient, "--encrypt"]
        else:
            raise ValueError(
                "Invalid action: '%s'. Expected 'encrypt' or 'decrypt'." % action
            )
        if output_filepath:
            # Specify output file path for the encrypted/decrypted file.
            options += ["--output", output_filepath]
        # Specify the file to encrypt/decrypt.
        if input_filepath:
            options += [input_filepath]
        # Ensure that the "check" argument is set to True by default.
        kwargs.setdefault("check", True)
        Command.__init__(self, gpg, args=options, **kwargs)


def shell_quote(arg):
    """
    Quote a string argument to be safely included in a shell command line.

    :param str arg: The script argument
    :return: The argument quoted
    """

    # This is an excerpt of the Bash manual page, and the same applies for
    # every Posix compliant shell:
    #
    #     A  non-quoted backslash (\) is the escape character.  It preserves
    #     the literal value of the next character that follows, with the
    #     exception of <newline>.  If a \<newline> pair appears, and the
    #     backslash is not itself quoted, the \<newline> is treated as a
    #     line continuation (that is, it is removed  from  the  input
    #     stream  and  effectively ignored).
    #
    #     Enclosing characters in single quotes preserves the literal value
    #     of each character within the quotes.  A single quote may not occur
    #     between single quotes, even when pre-ceded by a backslash.
    #
    # This means that, as long as the original string doesn't contain any
    # apostrophe character, it can be safely included between single quotes.
    #
    # If a single quote is contained in the string, we must terminate the
    # string with a quote, insert an apostrophe character escaping it with
    # a backslash, and then start another string using a quote character.

    assert arg is not None
    if arg == "|":
        return arg
    return "'%s'" % arg.replace("'", "'\\''")


def full_command_quote(command, args=None):
    """
    Produce a command with quoted arguments

    :param str command: the command to be executed
    :param list[str] args: the command arguments
    :rtype: str
    """
    if args is not None and len(args) > 0:
        return "%s %s" % (command, " ".join([shell_quote(arg) for arg in args]))
    else:
        return command



# -*- coding: utf-8 -*-
# © Copyright EnterpriseDB UK Limited 2018-2025
#
# This file is part of Barman.
#
# Barman is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Barman is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Barman.  If not, see <http://www.gnu.org/licenses/>

import json
import logging
import math
import os
import shutil
from datetime import datetime
from io import RawIOBase

from barman.clients.cloud_compression import decompress_to_file
from barman.cloud import (
    DEFAULT_DELIMITER,
    CloudInterface,
    CloudProviderError,
    CloudSnapshotInterface,
    DecompressingStreamingIO,
    SnapshotMetadata,
    SnapshotsInfo,
    VolumeMetadata,
)
from barman.exceptions import (
    CommandException,
    SnapshotBackupException,
    SnapshotInstanceNotFoundException,
)

try:
    # Python 3.x
    from urllib.parse import urlencode, urlparse
except ImportError:
    # Python 2.x
    from urllib import urlencode

    from urlparse import urlparse

try:
    import boto3
    from boto3.s3.transfer import TransferConfig
    from botocore.config import Config
    from botocore.exceptions import ClientError, EndpointConnectionError
except ImportError:
    raise SystemExit("Missing required python module: boto3")


class StreamingBodyIO(RawIOBase):
    """
    Wrap a boto StreamingBody in the IOBase API.
    """

    def __init__(self, body):
        self.body = body

    def readable(self):
        return True

    def read(self, n=-1):
        n = None if n < 0 else n
        return self.body.read(n)


class S3CloudInterface(CloudInterface):
    # S3 multipart upload limitations
    # http://docs.aws.amazon.com/AmazonS3/latest/API/mpUploadUploadPart.html
    MAX_CHUNKS_PER_FILE = 10000
    MIN_CHUNK_SIZE = 5 << 20

    # S3 permit a maximum of 5TB per file
    # https://docs.aws.amazon.com/AmazonS3/latest/dev/UploadingObjects.html
    # This is a hard limit, while our upload procedure can go over the specified
    # MAX_ARCHIVE_SIZE - so we set a maximum of 1TB per file
    MAX_ARCHIVE_SIZE = 1 << 40

    MAX_DELETE_BATCH_SIZE = 1000

    # The minimum size for a file to be uploaded using multipart upload in upload_fileobj
    # 100MB is the AWS recommendation for when to start considering using multipart upload
    # https://docs.aws.amazon.com/AmazonS3/latest/userguide/mpuoverview.html
    MULTIPART_THRESHOLD = 104857600

    def __getstate__(self):
        state = self.__dict__.copy()
        # Remove boto3 client reference from the state as it cannot be pickled
        # in Python >= 3.8 and multiprocessing will pickle the object when the
        # worker processes are created.
        # The worker processes create their own boto3 sessions so do not need
        # the boto3 session from the parent process.
        del state["s3"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __init__(
        self,
        url,
        encryption=None,
        jobs=2,
        profile_name=None,
        endpoint_url=None,
        tags=None,
        delete_batch_size=None,
        read_timeout=None,
        sse_kms_key_id=None,
    ):
        """
        Create a new S3 interface given the S3 destination url and the profile
        name

        :param str url: Full URL of the cloud destination/source
        :param str|None encryption: Encryption type string
        :param int jobs: How many sub-processes to use for asynchronous
          uploading, defaults to 2.
        :param str profile_name: Amazon auth profile identifier
        :param str endpoint_url: override default endpoint detection strategy
          with this one
        :param int|None delete_batch_size: the maximum number of objects to be
          deleted in a single request
        :param int|None read_timeout: the time in seconds until a timeout is
          raised when waiting to read from a connection
        :param str|None sse_kms_key_id: the AWS KMS key ID that should be used
          for encrypting uploaded data in S3
        """
        super(S3CloudInterface, self).__init__(
            url=url,
            jobs=jobs,
            tags=tags,
            delete_batch_size=delete_batch_size,
        )
        self.profile_name = profile_name
        self.encryption = encryption
        self.endpoint_url = endpoint_url
        self.read_timeout = read_timeout
        self.sse_kms_key_id = sse_kms_key_id

        # Extract information from the destination URL
        parsed_url = urlparse(url)
        # If netloc is not present, the s3 url is badly formatted.
        if parsed_url.netloc == "" or parsed_url.scheme != "s3":
            raise ValueError("Invalid s3 URL address: %s" % url)
        self.bucket_name = parsed_url.netloc
        self.bucket_exists = None
        self.path = parsed_url.path.lstrip("/")

        # initialize the config object to be used in uploads
        self.config = TransferConfig(multipart_threshold=self.MULTIPART_THRESHOLD)

        # Build a session, so we can extract the correct resource
        self._reinit_session()

    def _reinit_session(self):
        """
        Create a new session
        """
        config_kwargs = {}
        if self.read_timeout is not None:
            config_kwargs["read_timeout"] = self.read_timeout
        config = Config(**config_kwargs)

        session = boto3.Session(profile_name=self.profile_name)
        self.s3 = session.resource("s3", endpoint_url=self.endpoint_url, config=config)

    @property
    def _extra_upload_args(self):
        """
        Return a dict containing ExtraArgs to be passed to certain boto3 calls

        Because some boto3 calls accept `ExtraArgs: {}` and others do not, we
        return a nested dict which can be expanded with `**` in the boto3 call.
        """
        additional_args = {}
        if self.encryption:
            additional_args["ServerSideEncryption"] = self.encryption
        if self.sse_kms_key_id:
            additional_args["SSEKMSKeyId"] = self.sse_kms_key_id
        return additional_args

    def test_connectivity(self):
        """
        Test AWS connectivity by trying to access a bucket
        """
        try:
            # We are not even interested in the existence of the bucket,
            # we just want to try if aws is reachable
            self.bucket_exists = self._check_bucket_existence()
            return True
        except EndpointConnectionError as exc:
            logging.error("Can't connect to cloud provider: %s", exc)
            return False

    def _check_bucket_existence(self):
        """
        Check cloud storage for the target bucket

        :return: True if the bucket exists, False otherwise
        :rtype: bool
        """
        try:
            # Search the bucket on s3
            self.s3.meta.client.head_bucket(Bucket=self.bucket_name)
            return True
        except ClientError as exc:
            # If a client error is thrown, then check the error code.
            # If code was 404, then the bucket does not exist
            error_code = exc.response["Error"]["Code"]
            if error_code == "404":
                return False
            # Otherwise there is nothing else to do than re-raise the original
            # exception
            raise

    def _create_bucket(self):
        """
        Create the bucket in cloud storage
        """
        # Get the current region from client.
        # Do not use session.region_name here because it may be None
        region = self.s3.meta.client.meta.region_name
        logging.info(
            "Bucket '%s' does not exist, creating it on region '%s'",
            self.bucket_name,
            region,
        )
        create_bucket_config = {
            "ACL": "private",
        }
        # The location constraint is required during bucket creation
        # for all regions outside of us-east-1. This constraint cannot
        # be specified in us-east-1; specifying it in this region
        # results in a failure, so we will only
        # add it if we are deploying outside of us-east-1.
        # See https://github.com/boto/boto3/issues/125
        if region != "us-east-1":
            create_bucket_config["CreateBucketConfiguration"] = {
                "LocationConstraint": region,
            }
        self.s3.Bucket(self.bucket_name).create(**create_bucket_config)

    def list_bucket(self, prefix="", delimiter=DEFAULT_DELIMITER):
        """
        List bucket content in a directory manner

        :param str prefix:
        :param str delimiter:
        :return: List of objects and dirs right under the prefix
        :rtype: List[str]
        """
        if prefix.startswith(delimiter):
            prefix = prefix.lstrip(delimiter)

        paginator = self.s3.meta.client.get_paginator("list_objects_v2")
        pages = paginator.paginate(
            Bucket=self.bucket_name, Prefix=prefix, Delimiter=delimiter
        )

        for page in pages:
            # List "folders"
            keys = page.get("CommonPrefixes")
            if keys is not None:
                for k in keys:
                    yield k.get("Prefix")

            # List "files"
            objects = page.get("Contents")
            if objects is not None:
                for o in objects:
                    yield o.get("Key")

    def download_file(self, key, dest_path, decompress):
        """
        Download a file from S3

        :param str key: The S3 key to download
        :param str dest_path: Where to put the destination file
        :param str|None decompress: Compression scheme to use for decompression
        """
        # Open the remote file
        obj = self.s3.Object(self.bucket_name, key)
        remote_file = obj.get()["Body"]

        # Write the dest file in binary mode
        with open(dest_path, "wb") as dest_file:
            # If the file is not compressed, just copy its content
            if decompress is None:
                shutil.copyfileobj(remote_file, dest_file)
                return

            decompress_to_file(remote_file, dest_file, decompress)

    def remote_open(self, key, decompressor=None):
        """
        Open a remote S3 object and returns a readable stream

        :param str key: The key identifying the object to open
        :param barman.clients.cloud_compression.ChunkedCompressor decompressor:
          A ChunkedCompressor object which will be used to decompress chunks of bytes
          as they are read from the stream
        :return: A file-like object from which the stream can be read or None if
          the key does not exist
        """
        try:
            obj = self.s3.Object(self.bucket_name, key)
            resp = StreamingBodyIO(obj.get()["Body"])
            if decompressor:
                return DecompressingStreamingIO(resp, decompressor)
            else:
                return resp
        except ClientError as exc:
            error_code = exc.response["Error"]["Code"]
            if error_code == "NoSuchKey":
                return None
            else:
                raise

    def upload_fileobj(self, fileobj, key, override_tags=None):
        """
        Synchronously upload the content of a file-like object to a cloud key

        :param fileobj IOBase: File-like object to upload
        :param str key: The key to identify the uploaded object
        :param List[tuple] override_tags: List of k,v tuples which should override any
          tags already defined in the cloud interface
        """
        extra_args = self._extra_upload_args.copy()
        tags = override_tags or self.tags
        if tags is not None:
            extra_args["Tagging"] = urlencode(tags)
        self.s3.meta.client.upload_fileobj(
            Fileobj=fileobj,
            Bucket=self.bucket_name,
            Key=key,
            ExtraArgs=extra_args,
            Config=self.config,
        )

    def create_multipart_upload(self, key):
        """
        Create a new multipart upload

        :param key: The key to use in the cloud service
        :return: The multipart upload handle
        :rtype: dict[str, str]
        """
        extra_args = self._extra_upload_args.copy()
        if self.tags is not None:
            extra_args["Tagging"] = urlencode(self.tags)
        return self.s3.meta.client.create_multipart_upload(
            Bucket=self.bucket_name, Key=key, **extra_args
        )

    def _upload_part(self, upload_metadata, key, body, part_number):
        """
        Upload a part into this multipart upload

        :param dict upload_metadata: The multipart upload handle
        :param str key: The key to use in the cloud service
        :param object body: A stream-like object to upload
        :param int part_number: Part number, starting from 1
        :return: The part handle
        :rtype: dict[str, None|str]
        """
        part = self.s3.meta.client.upload_part(
            Body=body,
            Bucket=self.bucket_name,
            Key=key,
            UploadId=upload_metadata["UploadId"],
            PartNumber=part_number,
        )
        return {
            "PartNumber": part_number,
            "ETag": part["ETag"],
        }

    def _complete_multipart_upload(self, upload_metadata, key, parts):
        """
        Finish a certain multipart upload

        :param dict upload_metadata:  The multipart upload handle
        :param str key: The key to use in the cloud service
        :param parts: The list of parts composing the multipart upload
        """
        self.s3.meta.client.complete_multipart_upload(
            Bucket=self.bucket_name,
            Key=key,
            UploadId=upload_metadata["UploadId"],
            MultipartUpload={"Parts": parts},
        )

    def _abort_multipart_upload(self, upload_metadata, key):
        """
        Abort a certain multipart upload

        :param dict upload_metadata:  The multipart upload handle
        :param str key: The key to use in the cloud service
        """
        self.s3.meta.client.abort_multipart_upload(
            Bucket=self.bucket_name, Key=key, UploadId=upload_metadata["UploadId"]
        )

    def _delete_objects_batch(self, paths):
        """
        Delete the objects at the specified paths

        :param List[str] paths:
        """
        super(S3CloudInterface, self)._delete_objects_batch(paths)

        resp = self.s3.meta.client.delete_objects(
            Bucket=self.bucket_name,
            Delete={
                "Objects": [{"Key": path} for path in paths],
                "Quiet": True,
            },
        )
        if "Errors" in resp:
            for error_dict in resp["Errors"]:
                logging.error(
                    'Deletion of object %s failed with error code: "%s", message: "%s"'
                    % (error_dict["Key"], error_dict["Code"], error_dict["Message"])
                )
            raise CloudProviderError()

    def get_prefixes(self, prefix):
        """
        Return only the common prefixes under the supplied prefix.

        :param str prefix: The object key prefix under which the common prefixes
            will be found.
        :rtype: Iterator[str]
        :return: A list of unique prefixes immediately under the supplied prefix.
        """
        for wal_prefix in self.list_bucket(prefix + "/", delimiter="/"):
            if wal_prefix.endswith("/"):
                yield wal_prefix

    def delete_under_prefix(self, prefix):
        """
        Delete all objects under the specified prefix.

        :param str prefix: The object key prefix under which all objects should be
            deleted.
        """
        if len(prefix) == 0 or prefix == "/" or not prefix.endswith("/"):
            raise ValueError(
                "Deleting all objects under prefix %s is not allowed" % prefix
            )
        bucket = self.s3.Bucket(self.bucket_name)
        for resp in bucket.objects.filter(Prefix=prefix).delete():
            response_metadata = resp["ResponseMetadata"]
            if response_metadata["HTTPStatusCode"] != 200:
                logging.error(
                    'Deletion of objects under %s failed with error code: "%s"'
                    % (prefix, response_metadata["HTTPStatusCode"])
                )
                raise CloudProviderError()


class AwsCloudSnapshotInterface(CloudSnapshotInterface):
    """
    Implementation of CloudSnapshotInterface for EBS snapshots as implemented in AWS
    as documented at:

        https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/ebs-creating-snapshot.html
    """

    def __init__(
        self,
        profile_name=None,
        region=None,
        await_snapshots_timeout=3600,
        lock_mode=None,
        lock_duration=None,
        lock_cool_off_period=None,
        lock_expiration_date=None,
        tags=None,
    ):
        """
        Creates the client necessary for creating and managing snapshots.

        :param str profile_name: AWS auth profile identifier.
        :param str region: The AWS region in which snapshot resources are located.
        :param int await_snapshots_timeout: The maximum time in seconds to wait for
            snapshots to complete.
        :param str lock_mode: The lock mode to apply to the snapshot.
        :param int lock_duration: The duration (in days) for which the snapshot
            should be locked.
        :param int lock_cool_off_period: The cool-off period (in hours) for the snapshot.
        :param str lock_expiration_date: The expiration date for the snapshot in the format
            ``YYYY-MM-DDThh:mm:ss.sssZ``.
        :param List[Tuple[str, str]] tags: Key value pairs for tags to be applied.
        """

        self.session = boto3.Session(profile_name=profile_name)
        # If a specific region was provided then this overrides any region which may be
        # defined in the profile
        self.region = region or self.session.region_name
        self.ec2_client = self.session.client("ec2", region_name=self.region)
        self.await_snapshots_timeout = await_snapshots_timeout
        self.tags = tags
        self.lock_mode = lock_mode
        self.lock_duration = lock_duration
        self.lock_cool_off_period = lock_cool_off_period
        self.lock_expiration_date = lock_expiration_date

    def _get_waiter_config(self):
        delay = 15
        # Use ceil so that we always wait for at least the specified timeout
        max_attempts = math.ceil(self.await_snapshots_timeout / delay)
        return {
            "Delay": delay,
            # Ensure we always try waiting at least once
            "MaxAttempts": max(max_attempts, 1),
        }

    def _get_instance_metadata(self, instance_identifier):
        """
        Retrieve the boto3 describe_instances metadata for the specified instance.

        The supplied instance_identifier can be either an AWS instance ID or a name.
        If an instance ID is supplied then this function will look it up directly. If
        a name is supplied then the `tag:Name` filter will be used to query the AWS
        API for instances with the matching `Name` tag.

        :param str instance_identifier: The instance ID or name of the VM instance.
        :rtype: dict
        :return: A dict containing the describe_instances metadata for the specified
            VM instance.
        """
        # Consider all states other than `terminated` as valid instances
        allowed_states = ["pending", "running", "shutting-down", "stopping", "stopped"]
        # If the identifier looks like an instance ID then we attempt to look it up
        resp = None
        if instance_identifier.startswith("i-"):
            try:
                resp = self.ec2_client.describe_instances(
                    InstanceIds=[instance_identifier],
                    Filters=[
                        {"Name": "instance-state-name", "Values": allowed_states},
                    ],
                )
            except ClientError as exc:
                error_code = exc.response["Error"]["Code"]
                # If we have a malformed instance ID then continue and treat it
                # like a name, otherwise re-raise the original error
                if error_code != "InvalidInstanceID.Malformed":
                    raise
        # If we do not have a response then try looking up by name
        if resp is None:
            resp = self.ec2_client.describe_instances(
                Filters=[
                    {"Name": "tag:Name", "Values": [instance_identifier]},
                    {"Name": "instance-state-name", "Values": allowed_states},
                ]
            )
        # Check for non-unique reservations and instances before returning the instance
        # because tag uniqueness is not a thing
        reservations = resp["Reservations"]
        if len(reservations) == 1:
            if len(reservations[0]["Instances"]) == 1:
                return reservations[0]["Instances"][0]
            elif len(reservations[0]["Instances"]) > 1:
                raise CloudProviderError(
                    "Cannot find a unique EC2 instance matching {}".format(
                        instance_identifier
                    )
                )
        elif len(reservations) > 1:
            raise CloudProviderError(
                "Cannot find a unique EC2 reservation containing instance {}".format(
                    instance_identifier
                )
            )
        raise SnapshotInstanceNotFoundException(
            "Cannot find instance {}".format(instance_identifier)
        )

    def _has_tag(self, resource, tag_name, tag_value):
        """
        Determine whether the resource metadata contains a specified tag.

        :param dict resource: Metadata describing an AWS resource.
        :parma str tag_name: The name of the tag to be checked.
        :param str tag_value: The value of the tag to be checked.
        :rtype: bool
        :return: True if a tag with the specified name and value was found, False
            otherwise.
        """
        if "Tags" in resource:
            for tag in resource["Tags"]:
                if tag["Key"] == tag_name and tag["Value"] == tag_value:
                    return True
        return False

    def _lookup_volume(self, attached_volumes, volume_identifier):
        """
        Searches a supplied list of describe_volumes metadata for the specified volume.

        :param list[dict] attached_volumes: A list of volumes in the format provided by
            the boto3 describe_volumes function.
        :param str volume_identifier: The volume ID or name of the volume to be looked
            up.
        :rtype: dict|None
        :return: describe_volume metadata for the volume matching the supplied
            identifier.
        """
        # Check whether volume_identifier matches a VolumeId
        matching_volumes = [
            volume
            for volume in attached_volumes
            if volume["VolumeId"] == volume_identifier
        ]
        # If we do not have a match, try again but search for a matching Name tag
        if not matching_volumes:
            matching_volumes = [
                volume
                for volume in attached_volumes
                if self._has_tag(volume, "Name", volume_identifier)
            ]
        # If there is more than one matching volume then it's an error condition
        if len(matching_volumes) > 1:
            raise CloudProviderError(
                "Duplicate volumes found matching {}: {}".format(
                    volume_identifier,
                    ", ".join(v["VolumeId"] for v in matching_volumes),
                )
            )
        # If no matching volumes were found then return None - it is up to the calling
        # code to decide if this is an error
        elif len(matching_volumes) == 0:
            return None
        # Otherwise, we found exactly one matching volume and return its metadata
        else:
            return matching_volumes[0]

    def _get_requested_volumes(self, instance_metadata, disks=None):
        """
        Fetch describe_volumes metadata for disks attached to a specified VM instance.

        Queries the AWS API for metadata describing the volumes attached to the
        instance described in instance_metadata.

        If `disks` is specified then metadata is only returned for the volumes that are
        included in the list and attached to the instance. Volumes which are requested
        in the `disks` list but not attached to the instance are not included in the
        response - it is up to calling code to decide whether this is an error
        condition.

        Entries in `disks` can be either volume IDs or names. The value provided for
        each volume will be included in the response under the key `identifier`.

        If `disks` is not provided then every non-root volume attached to the instance
        will be included in the response.

        :param dict instance_metadata: A dict containing the describe_instances metadata
            for a VM instance.
        :param list[str] disks: A list of volume IDs or volume names. If specified then
            only volumes in this list which are attached to the instance described by
            instance_metadata will be included in the response.
        :rtype: list[dict[str,str|dict]]
        :return: A list of dicts containing identifiers and describe_volumes metadata
            for the requested volumes.
        """
        # Pre-fetch the describe_volumes output for all volumes attached to the instance
        attached_volumes = self.ec2_client.describe_volumes(
            Filters=[
                {
                    "Name": "attachment.instance-id",
                    "Values": [instance_metadata["InstanceId"]],
                },
            ]
        )["Volumes"]
        # If disks is None then use a list of all Ebs volumes attached to the instance
        requested_volumes = []
        if disks is None:
            disks = [
                device["Ebs"]["VolumeId"]
                for device in instance_metadata["BlockDeviceMappings"]
                if "Ebs" in device
            ]
        # For each requested volume, look it up in the describe_volumes output using
        # _lookup_volume which will handle both volume IDs and volume names
        for volume_identifier in disks:
            volume = self._lookup_volume(attached_volumes, volume_identifier)
            if volume is not None:
                attachment_metadata = None
                for attachment in volume["Attachments"]:
                    if attachment["InstanceId"] == instance_metadata["InstanceId"]:
                        attachment_metadata = attachment
                        break
                if attachment_metadata is not None:
                    # Ignore the root volume
                    if (
                        attachment_metadata["Device"]
                        == instance_metadata["RootDeviceName"]
                    ):
                        continue
                    snapshot_id = None
                    if "SnapshotId" in volume and volume["SnapshotId"] != "":
                        snapshot_id = volume["SnapshotId"]
                    requested_volumes.append(
                        {
                            "identifier": volume_identifier,
                            "attachment_metadata": attachment_metadata,
                            "source_snapshot": snapshot_id,
                        }
                    )
        return requested_volumes

    def _create_snapshot(self, backup_info, volume_name, volume_id):
        """
        Create a snapshot of an EBS volume in AWS.

        Unlike its counterparts in AzureCloudSnapshotInterface and
        GcpCloudSnapshotInterface, this function does not wait for the snapshot to
        enter a successful completed state and instead relies on the calling code
        to perform any necessary waiting.

        :param barman.infofile.LocalBackupInfo backup_info: Backup information.
        :param str volume_name: The user-supplied identifier for the volume. Used
            when creating the snapshot name.
        :param str volume_id: The AWS volume ID. Used when calling the AWS API to
            create the snapshot.
        :rtype: (str, dict)
        :return: The snapshot name and the snapshot metadata returned by AWS.
        """
        snapshot_name = "%s-%s" % (
            volume_name,
            backup_info.backup_id.lower(),
        )
        logging.info(
            "Taking snapshot '%s' of disk '%s' (%s)",
            snapshot_name,
            volume_name,
            volume_id,
        )
        tags = [
            {"Key": "Name", "Value": snapshot_name},
        ]

        if self.tags is not None:
            for key, value in self.tags:
                tags.append({"Key": key, "Value": value})

        resp = self.ec2_client.create_snapshot(
            TagSpecifications=[
                {
                    "ResourceType": "snapshot",
                    "Tags": tags,
                }
            ],
            VolumeId=volume_id,
        )

        if resp["State"] == "error":
            raise CloudProviderError(
                "Snapshot '{}' failed: {}".format(snapshot_name, resp)
            )

        return snapshot_name, resp

    def take_snapshot_backup(self, backup_info, instance_identifier, volumes):
        """
        Take a snapshot backup for the named instance.

        Creates a snapshot for each named disk and saves the required metadata
        to backup_info.snapshots_info as an AwsSnapshotsInfo object.

        :param barman.infofile.LocalBackupInfo backup_info: Backup information.
        :param str instance_identifier: The instance ID or name of the VM instance to
            which the disks to be backed up are attached.
        :param dict[str,barman.cloud_providers.aws_s3.AwsVolumeMetadata] volumes:
            Metadata describing the volumes to be backed up.
        """
        instance_metadata = self._get_instance_metadata(instance_identifier)
        attachment_metadata = instance_metadata["BlockDeviceMappings"]
        snapshots = []
        for volume_identifier, volume_metadata in volumes.items():
            attached_volumes = [
                v
                for v in attachment_metadata
                if v["Ebs"]["VolumeId"] == volume_metadata.id
            ]
            if len(attached_volumes) == 0:
                raise SnapshotBackupException(
                    "Disk %s not attached to instance %s"
                    % (volume_identifier, instance_identifier)
                )
            assert len(attached_volumes) == 1

            snapshot_name, snapshot_resp = self._create_snapshot(
                backup_info, volume_identifier, volume_metadata.id
            )
            # Apply lock on snapshot if lock mode is specified
            if self.lock_mode:
                self._lock_snapshot(
                    snapshot_resp["SnapshotId"],
                    self.lock_mode,
                    self.lock_duration,
                    self.lock_cool_off_period,
                    self.lock_expiration_date,
                )

            snapshots.append(
                AwsSnapshotMetadata(
                    snapshot_id=snapshot_resp["SnapshotId"],
                    snapshot_name=snapshot_name,
                    snapshot_lock_mode=self.lock_mode,
                    device_name=attached_volumes[0]["DeviceName"],
                    mount_options=volume_metadata.mount_options,
                    mount_point=volume_metadata.mount_point,
                )
            )

        # Await completion of all snapshots using a boto3 waiter. This will call
        # `describe_snapshots` every 15 seconds until all snapshot IDs are in a
        # successful state. If the successful state is not reached after the maximum
        # number of attempts (default: 40) then a WaiterError is raised.
        snapshot_ids = [snapshot.identifier for snapshot in snapshots]
        logging.info("Waiting for completion of snapshots: %s", ", ".join(snapshot_ids))
        waiter = self.ec2_client.get_waiter("snapshot_completed")
        waiter.wait(
            SnapshotIds=snapshot_ids,
            WaiterConfig=self._get_waiter_config(),
        )
        backup_info.snapshots_info = AwsSnapshotsInfo(
            snapshots=snapshots,
            region=self.region,
            # All snapshots will have the same OwnerId so we get it from the last
            # snapshot response.
            account_id=snapshot_resp["OwnerId"],
        )

    def _lock_snapshot(
        self,
        snapshot_id,
        lock_mode,
        lock_duration,
        lock_cool_off_period,
        lock_expiration_date,
    ):
        lock_snapshot_default_args = {"LockMode": lock_mode, "SnapshotId": snapshot_id}

        if lock_duration:
            lock_snapshot_default_args["LockDuration"] = lock_duration

        if lock_cool_off_period:
            lock_snapshot_default_args["CoolOffPeriod"] = lock_cool_off_period

        if lock_expiration_date:
            lock_snapshot_default_args["ExpirationDate"] = lock_expiration_date

        resp = self.ec2_client.lock_snapshot(**lock_snapshot_default_args)

        _output = {}
        for key, value in resp.items():
            if key != "ResponseMetadata":
                if isinstance(value, datetime):
                    value = value.isoformat()
                _output[key] = value

        logging.info("Snapshot locked: \n%s" % json.dumps(_output, indent=4))

    def _delete_snapshot(self, snapshot_id):
        """
        Delete the specified snapshot.

        :param str snapshot_id: The ID of the snapshot to be deleted.
        """
        try:
            self.ec2_client.delete_snapshot(SnapshotId=snapshot_id)
        except ClientError as exc:
            error_code = exc.response["Error"]["Code"]
            # If the snapshot could not be found then deletion is considered successful
            # otherwise we raise a CloudProviderError
            if error_code == "InvalidSnapshot.NotFound":
                logging.warning("Snapshot {} could not be found".format(snapshot_id))
            elif error_code == "SnapshotLocked":
                raise SystemExit(
                    "Locked snapshot: %s.\n"
                    "Before deleting a snapshot, please ensure that it is not locked "
                    "or that the lock has expired." % snapshot_id,
                )
            else:
                raise CloudProviderError(
                    "Deletion of snapshot %s failed with error code %s: %s"
                    % (snapshot_id, error_code, exc.response["Error"])
                )
        logging.info("Snapshot %s deleted", snapshot_id)

    def delete_snapshot_backup(self, backup_info):
        """
        Delete all snapshots for the supplied backup.

        :param barman.infofile.LocalBackupInfo backup_info: Backup information.
        """
        for snapshot in backup_info.snapshots_info.snapshots:
            logging.info(
                "Deleting snapshot '%s' for backup %s",
                snapshot.identifier,
                backup_info.backup_id,
            )
            self._delete_snapshot(snapshot.identifier)

    def get_attached_volumes(
        self, instance_identifier, disks=None, fail_on_missing=True
    ):
        """
        Returns metadata for the non-root volumes attached to this instance.

        Queries AWS for metadata relating to the volumes attached to the named instance
        and returns a dict of `VolumeMetadata` objects, keyed by volume identifier.

        The volume identifier will be either:
        - The value supplied in the disks parameter, which can be either the AWS
          assigned volume ID or a name which corresponds to a unique `Name` tag assigned
          to a volume.
        - The AWS assigned volume ID, if the disks parameter is unused.

        If the optional disks parameter is supplied then this method returns metadata
        for the disks in the supplied list only. If fail_on_missing is set to True then
        a SnapshotBackupException is raised if any of the supplied disks are not found
        to be attached to the instance.

        If the disks parameter is not supplied then this method returns a
        VolumeMetadata object for every non-root disk attached to this instance.

        :param str instance_identifier: Either an instance ID or the name of the VM
            instance to which the disks are attached.
        :param list[str]|None disks: A list containing either the volume IDs or names
            of disks backed up.
        :param bool fail_on_missing: Fail with a SnapshotBackupException if any
            specified disks are not attached to the instance.
        :rtype: dict[str, VolumeMetadata]
        :return: A dict where the key is the volume identifier and the value is the
            device path for that disk on the specified instance.
        """
        instance_metadata = self._get_instance_metadata(instance_identifier)
        requested_volumes = self._get_requested_volumes(instance_metadata, disks)

        attached_volumes = {}
        for requested_volume in requested_volumes:
            attached_volumes[requested_volume["identifier"]] = AwsVolumeMetadata(
                requested_volume["attachment_metadata"],
                virtualization_type=instance_metadata["VirtualizationType"],
                source_snapshot=requested_volume["source_snapshot"],
            )

        if disks is not None and fail_on_missing:
            unattached_volumes = []
            for disk_identifier in disks:
                if disk_identifier not in attached_volumes:
                    unattached_volumes.append(disk_identifier)
            if len(unattached_volumes) > 0:
                raise SnapshotBackupException(
                    "Disks not attached to instance {}: {}".format(
                        instance_identifier, ", ".join(unattached_volumes)
                    )
                )
        return attached_volumes

    def instance_exists(self, instance_identifier):
        """
        Determine whether the instance exists.

        :param str instance_identifier: A string identifying the VM instance to be
            checked. Can be either an instance ID or a name. If a name is provided
            it is expected to match the value of a `Name` tag for a single EC2
            instance.
        :rtype: bool
        :return: True if the named instance exists, False otherwise.
        """
        try:
            self._get_instance_metadata(instance_identifier)
        except SnapshotInstanceNotFoundException:
            return False
        return True


class AwsVolumeMetadata(VolumeMetadata):
    """
    Specialization of VolumeMetadata for AWS EBS volumes.

    This class uses the device name obtained from the AWS API together with the
    virtualization type of the VM to which it is attached in order to resolve the
    mount point and mount options for the volume.
    """

    def __init__(
        self, attachment_metadata=None, virtualization_type=None, source_snapshot=None
    ):
        """
        Creates an AwsVolumeMetadata instance using metadata obtained from the AWS API.

        :param dict attachment_metadata: An `Attachments` entry in the describe_volumes
            metadata for this volume.
        :param str virtualization_type: The type of virtualzation used by the VM to
            which this volume is attached - either "hvm" or "paravirtual".
        :param str source_snapshot: The snapshot ID of the source snapshot from which
            volume was created.
        """
        super(AwsVolumeMetadata, self).__init__()
        # The `id` property is used to store the volume ID so that we always have a
        # reference to the canonical ID of the volume. This is essential when creating
        # snapshots via the AWS API.
        self.id = None
        self._device_name = None
        self._virtualization_type = virtualization_type
        self._source_snapshot = source_snapshot
        if attachment_metadata:
            if "Device" in attachment_metadata:
                self._device_name = attachment_metadata["Device"]
            if "VolumeId" in attachment_metadata:
                self.id = attachment_metadata["VolumeId"]

    def resolve_mounted_volume(self, cmd):
        """
        Resolve the mount point and mount options using shell commands.

        Uses `findmnt` to locate the mount point and options for the volume.

        Search for the system's available EBS volume device names and IDs from the
        `/sys/block` directory in the sysfs virtual filesystem that provides
        information and configuration options for block devices on the system.

        For Nitro instances, the NVMe device name can change depending on the order
        in which the devices respond during instance boot. AWS recommends using the
        EBS volume ID and the mount point for consistent device identification.

        Other instance types do not have a volume ID associated with them. Instead,
        we identify the device from the device name or an alternative name obtained
        by renaming the device name.

        This is based on information provided by AWS about device identification and
        renaming in EC2:
        - https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/device_naming.html
        - https://docs.aws.amazon.com/ebs/latest/userguide/identify-nvme-ebs-device.html

        :param UnixLocalCommand cmd: An object which can be used to run shell commands
            on a local (or remote, via the UnixRemoteCommand subclass) instance.
        """
        if self._device_name is None:
            raise SnapshotBackupException(
                "Cannot resolve mounted volume: device name unknown"
            )

        name = self._device_name
        # Check if it is a root volume.
        if self._device_name in ("/dev/sda1", "/dev/xvda"):
            raise SnapshotBackupException(
                "'%s' is a root volume. EBS volumes envolved in the snapshot "
                "backup must be non-root." % self._device_name
            )

        # Virtualization options: "paravirtual" / "hvm".
        # HVM is more common in AWS EC2 environments so we use '/dev/xvd' as
        # "default" new prefix.
        new_prefix = "/dev/xvd"
        if self._virtualization_type == "paravirtual":
            new_prefix = "/dev/hd"

        # OS sysfs virtual filesystem blocks path.
        sys_block_path = "/sys/block"
        # Get block names from the OS sysfs virtual filesystem.
        blocks = os.listdir(sys_block_path)
        device_prefix = "/dev/sd"
        for block in blocks:
            device = "/dev/%s" % block
            # Check for non Nitro instances using the block name directly.
            if name == device:
                break
            # Check for non Nitro instances using a renamed version of the device name.
            # AWS renames the device names attached to an instance automatically, so we
            # must check for this possibility as well.
            if name.startswith(device_prefix):
                renamed = name.replace(device_prefix, new_prefix)
                if renamed == device:
                    name = renamed
                    break
            # Check for Nitro instances. We need to retrieve the volume id from the file
            # which can be found at the `/sys/block/{block}/device/serial` path and use
            # itto find the correct device name.
            serial_number_path = "%s/%s/device/serial" % (sys_block_path, block)
            if "nvme" in device and os.path.isfile(serial_number_path):
                with open(serial_number_path, "r") as file:
                    serial_number = file.readline().strip()
                # ``self.id`` format: `vol-{HASH}`.
                vol_id = serial_number.replace("vol", "vol-")
                # If the volumes match, device ``name`` is found.
                if vol_id == self.id:
                    name = device
                    break
        try:
            # The NVMe device name format can vary depending on whether the EBS volume
            # was attached during or after the instance launch. NVMe device names for
            # volumes attached after instance launch include the `/dev/` prefix, while
            # NVMe device names for volumes attached during instance launch do not
            # include the `/dev/` prefix.
            for dev_name in [name, name.lstrip("/dev/")]:
                mount_point, mount_options = cmd.findmnt(dev_name)
                if mount_point is not None:
                    self._mount_point = mount_point
                    self._mount_options = mount_options
                    return
        except CommandException as e:
            raise SnapshotBackupException(
                "Error finding mount point for device path %s: %s"
                % (self._device_name, e)
            )
        raise SnapshotBackupException(
            "Could not find device %s at any mount point" % self._device_name
        )

    @property
    def source_snapshot(self):
        """
        An identifier which can reference the snapshot via the cloud provider.

        :rtype: str
        :return: The snapshot ID
        """
        return self._source_snapshot


class AwsSnapshotMetadata(SnapshotMetadata):
    """
    Specialization of SnapshotMetadata for AWS EBS snapshots.

    Stores the device_name, snapshot_id, snapshot_name and snapshot_lock_mode in the provider-specific
    field.
    """

    _provider_fields = (
        "device_name",
        "snapshot_id",
        "snapshot_name",
        "snapshot_lock_mode",
    )

    def __init__(
        self,
        mount_options=None,
        mount_point=None,
        device_name=None,
        snapshot_id=None,
        snapshot_name=None,
        snapshot_lock_mode=None,
    ):
        """
        Constructor saves additional metadata for AWS snapshots.

        :param str mount_options: The mount options used for the source disk at the
            time of the backup.
        :param str mount_point: The mount point of the source disk at the time of
            the backup.
        :param str device_name: The device name used in the AWS API.
        :param str snapshot_id: The snapshot ID used in the AWS API.
        :param str snapshot_name: The snapshot name stored in the `Name` tag.
        :param str snapshot_lock_mode: The mode with which the snapshot has been locked
            (``governance`` or ``compliance``), if set.
        :param str project: The AWS project name.
        """
        super(AwsSnapshotMetadata, self).__init__(mount_options, mount_point)
        self.device_name = device_name
        self.snapshot_id = snapshot_id
        self.snapshot_name = snapshot_name
        self.snapshot_lock_mode = snapshot_lock_mode

    @property
    def identifier(self):
        """
        An identifier which can reference the snapshot via the cloud provider.

        :rtype: str
        :return: The snapshot ID.
        """
        return self.snapshot_id


class AwsSnapshotsInfo(SnapshotsInfo):
    """
    Represents the snapshots_info field for AWS EBS snapshots.
    """

    _provider_fields = (
        "account_id",
        "region",
    )
    _snapshot_metadata_cls = AwsSnapshotMetadata

    def __init__(self, snapshots=None, account_id=None, region=None):
        """
        Constructor saves the list of snapshots if it is provided.

        :param list[SnapshotMetadata] snapshots: A list of metadata objects for each
            snapshot.
        :param str account_id: The AWS account to which the snapshots belong, as
            reported by the `OwnerId` field in the snapshots metadata returned by AWS
            at snapshot creation time.
        :param str region: The AWS region in which snapshot resources are located.
        """
        super(AwsSnapshotsInfo, self).__init__(snapshots)
        self.provider = "aws"
        self.account_id = account_id
        self.region = region



# -*- coding: utf-8 -*-
# © Copyright EnterpriseDB UK Limited 2018-2025
#
# This file is part of Barman.
#
# Barman is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Barman is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Barman.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
from contextlib import closing

from barman.clients.cloud_cli import (
    GeneralErrorExit,
    NetworkErrorExit,
    OperationErrorExit,
    create_argument_parser,
)
from barman.cloud import CloudBackupCatalog, configure_logging
from barman.cloud_providers import get_cloud_interface
from barman.infofile import BackupInfo
from barman.utils import force_str


def main(args=None):
    """
    The main script entry point

    :param list[str] args: the raw arguments list. When not provided
        it defaults to sys.args[1:]
    """
    config = parse_arguments(args)
    configure_logging(config)

    try:
        cloud_interface = get_cloud_interface(config)

        with closing(cloud_interface):
            catalog = CloudBackupCatalog(
                cloud_interface=cloud_interface, server_name=config.server_name
            )

            if not cloud_interface.test_connectivity():
                raise NetworkErrorExit()
            # If test is requested, just exit after connectivity test
            elif config.test:
                raise SystemExit(0)

            if not cloud_interface.bucket_exists:
                logging.error("Bucket %s does not exist", cloud_interface.bucket_name)
                raise OperationErrorExit()

            backup_list = catalog.get_backup_list()

            # Output
            if config.format == "console":
                COLUMNS = "{:<20}{:<25}{:<30}{:<17}{:<20}"
                print(
                    COLUMNS.format(
                        "Backup ID",
                        "End Time",
                        "Begin Wal",
                        "Archival Status",
                        "Name",
                    )
                )
                for backup_id in sorted(backup_list):
                    item = backup_list[backup_id]
                    if item and item.status == BackupInfo.DONE:
                        keep_target = catalog.get_keep_target(item.backup_id)
                        keep_status = (
                            keep_target and "KEEP:%s" % keep_target.upper() or ""
                        )
                        print(
                            COLUMNS.format(
                                item.backup_id,
                                item.end_time.strftime("%Y-%m-%d %H:%M:%S"),
                                item.begin_wal,
                                keep_status,
                                item.backup_name or "",
                            )
                        )
            else:
                print(
                    json.dumps(
                        {
                            "backups_list": [
                                backup_list[backup_id].to_json()
                                for backup_id in sorted(backup_list)
                            ]
                        }
                    )
                )

    except Exception as exc:
        logging.error("Barman cloud backup list exception: %s", force_str(exc))
        logging.debug("Exception details:", exc_info=exc)
        raise GeneralErrorExit()


def parse_arguments(args=None):
    """
    Parse command line arguments

    :return: The options parsed
    """

    parser, _, _ = create_argument_parser(
        description="This script can be used to list backups "
        "made with barman-cloud-backup command. "
        "Currently AWS S3, Azure Blob Storage and Google Cloud Storage are supported.",
    )

    parser.add_argument(
        "--format",
        default="console",
        help="Output format (console or json). Default console.",
    )
    return parser.parse_args(args=args)


if __name__ == "__main__":
    main()
